-- // Code by SoundOfSpouting#6980 (UID: 151149148639330304)

STATE_X, STATE_O, STATE_N = "X", "O", "."

function entry(str) return serialize(best(parse(str), STATE_X, STATE_O)) end

function parse(str)
  local board, x, y = {{},{},{}}, 1, 1
  for i=1,#str do
    local char = str:sub(i,i)
    if char == "\n" then y, x = y + 1, 1 else board[y][x], x = char, x + 1 end
  end
  return board
end

-- Returns the application of the first optimal move of the given position.
-- Notice: Does not care about the runtime of the game. The only criteria considered is the favorability of the final position.
function best(board, turn, next)
  local nboard, res = nil, winner(board)
  if res ~= STATE_N then return nil, res end
  local t = moves(board)
  for _, move in ipairs(t) do
    local x, y, _nboard = move.x, move.y, copy(board)
    _nboard[y][x] = turn

    local _, _res = best(_nboard, next, turn)
    if _res == turn then return _nboard, _res
    elseif nboard == nil or res == next and _res == STATE_N then nboard, res = _nboard, _res end
  end
  return nboard, res
end

function winner(board)
  for y=1,3 do if board[y][1] == board[y][2] and board[y][2] == board[y][3] or board[1][y] == board[2][y] and board[2][y] == board[3][y] then return board[y][y] end end
  if board[1][1] == board[2][2] and board[2][2] == board[3][3] or board[1][3] == board[2][2] and board[2][2] == board[3][1] then return board[2][2] end
  return STATE_N
end

function moves(board)
  local t = {}
  for y=1,3 do for x=1,3 do if board[y][x] == STATE_N then table.insert(t, {x = x, y = y}) end end end
  return t
end

function copy(board)
  _board = {{},{},{}}
  for y=1,3 do for x=1,3 do _board[y][x] = board[y][x] end end
  return _board
end

function serialize(board)
  local str = ""
  for y=1,3 do
    for x=1,3 do str = str .. board[y][x] end
    str = str .. "\n"
  end
  return str
end

-- Notice: Calculates future optimal moves repeatedly.
function play(str, turn, next, count)
  board = parse(str)
  print(serialize(board))
  for i=1,count do
    board, turn, next = best(board, turn, next), next, turn
    print(serialize(board))
    if winner(board) ~= STATE_N then break end
  end
  print(({[STATE_X] = "First player", [STATE_O] = "Second player", [STATE_N] = "Nobody"})[winner(board)] .. " wins!")
end
