module Win (precompute, form, pre) where

import Data.Foldable (maximumBy)
import Data.Function (on)
import Data.List (nub)
import Data.Maybe (fromMaybe)
import Language.Haskell.TH

won :: Char -> String -> Bool
won p [north_west, north, north_east, '\n', west, up, east, '\n', south_west, south, south_east] = line p [north_west, north, north_east, '\n', west, up, east, '\n', south_west, south, south_east] || line p [south_west, west, north_west, '\n', south, up, north, '\n', south_east, east, north_east]
  where
    line p (_ : _ : _ : '\n' : xs) | line p xs = True
    line p ('x' : 'x' : 'x' : _) | p == 'x' = True
    line p ('o' : 'o' : 'o' : _) | p == 'o' = True
    line p (_ : _ : 'x' : '\n' : _ : 'x' : _ : '\n' : 'x' : _) | p == 'x' = True
    line p (_ : _ : 'o' : '\n' : _ : 'o' : _ : '\n' : 'o' : _) | p == 'o' = True
    line _ _ = False

winning :: String -> Maybe (Maybe Char)
winning xs
  | won 'x' xs = Just (Just 'x')
  | won 'o' xs = Just (Just 'o')
  | '.' `notElem` xs = Just Nothing
winning _ = Nothing

turns :: Char -> String -> [String]
turns p xs = go p [] ([], xs)
  where
    go _ options (_, []) = map dezip options
      where
        dezip (hs, ts) = reverse hs ++ ts
    go p options (hs, h : ts) = go p (flex h options) (h : hs, ts)
      where
        flex '.' = ((hs, p : ts) :)
        flex c = id

form :: [String]
form = nub $ blank : concat [turns 'o' blank, turns 'x' blank >>= turns 'o', turns 'o' blank >>= turns 'x' >>= turns 'o']
  where
    blank = "...\n...\n..."

eval :: (Char, Char) -> String -> Either (Either Integer Integer) Integer
eval (p, o) xs = case winning xs of
  Just (Just best)
    | best == p -> Left $ Left 0
    | best == o -> Left $ Right 0
  Just Nothing -> Right 0
  Nothing -> next $ eval (o, p) $ move (o, p) xs
    where
      next (Right a) = Right $ pred a
      next (Left (Left a)) = Left $ Right $ pred a
      next (Left (Right a)) = Left $ Left $ pred a

move :: (Char, Char) -> String -> String
move (p, o) xs = maximumBy (test `on` eval (p, o)) $ turns p xs
  where
    test (Left (Left a)) (Left (Left b)) = compare b a
    test (Left (Right a)) (Left (Right b)) = compare a b
    test (Right a) (Right b) = compare b a
    test (Left (Left _)) _ = LT
    test (Left (Right _)) _ = GT
    test a b = compare EQ $ test b a

play :: (Char, Char) -> String -> [String]
play (p, o) xs = fromMaybe (xs : play (o, p) (move (p, o) xs)) $ [xs] <$ winning xs

pre :: String -> String
pre = move ('x', 'o')

precompute :: [String] -> DecsQ
precompute xs = pure [FunD (mkName "entry") $ clauses ++ [fallback]]
  where
    patterns = map (LitP . StringL) xs
    table = map (LitE . StringL . pre) xs
    clauses = zipWith (\body pattern -> Clause [pattern] (NormalB body) []) table patterns
    var = mkName "xs"
    body = AppE (VarE (mkName "pre")) (VarE var)
    fallback = Clause [VarP var] (NormalB body) []
