{-# LANGUAGE TupleSections #-}

import Control.Monad
import Data.Array.IArray
import Data.Function
import Data.Maybe
import Data.List

type Point = (Int, Int)
type Board = Array Point (Maybe Bool)

parseBoard :: String -> Board
parseBoard (a:b:c:'\n'
           :d:e:f:'\n'
           :g:h:i:_
           ) = listArray ((0, 0), (2, 2)) $ map charToCell [a, b, c, d, e, f, g, h, i]
  where
    charToCell 'x' = Just True
    charToCell 'o' = Just False
    charToCell '.' = Nothing

sayBoard :: Board -> String
sayBoard = map cellToChar . elems
  where
    cellToChar (Just True) = 'x'
    cellToChar (Just False) = 'o'
    cellToChar Nothing = '.'

mirror :: Board -> Maybe Point
mirror b = guard (b ! (1, 1) == Just True) >> (join . fmap msum . mapM hasMirror $ assocs b)
  where
    hasMirror (p@(x, y), c) | p == (1, 1) = Just Nothing
                            | otherwise =
      case c of
        Just False ->
          let i = (2-x, 2-y)
          in case b ! i of
            Just True -> Just Nothing
            Just False -> Nothing
            Nothing -> Just (Just i)
        _ -> Just Nothing

inARows :: Point -> [[Point]]
inARows (x, y) = map (x,) [0..2] : map (,y) [0..2] : diag 0 ++ diag 2
  where diag n = map ((,) <*> (n-)) [0..2] <$ (guard (x == n-y))

rowLength :: Board -> [Point] -> Int
rowLength b r
  | any ((==Just False) . (b !)) r = 0
  | otherwise = length $ filter ((==Just True) . (b !)) r

play :: Board -> Point
play b = case filter (isJust . snd) $ assocs b of
  [] -> (1, 1)
  [((x, y), _)]
    | odd (x + y) -> (y, x)
    | otherwise -> (0, 1)
  _ -> case mirror b of
    Just p -> p
    Nothing -> minimumBy (compare `on` (\p -> maximum . map (rowLength b) $ inARows p)) . map fst . filter (isNothing . snd) $ assocs b

entry :: String -> String
entry s = let b = parseBoard s in sayBoard $ b // [(play b, Just True)]

main = interact entry
